#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

from harkmw.logger import logger
from collections import defaultdict

import os
debug = (bool(os.environ['HARK_DEBUG'])
         ) if 'HARK_DEBUG' in os.environ else False


class Node:
    """ Base-class of Network, NativeNode, Node
    """

    def __init__(self):
        self.name = None
        self.type = None
        self.network = None
        self.count = -1
        self.inputs = defaultdict(lambda: None)
        self.outputs = defaultdict(lambda: [])
        self.input_counts = {}
        self.output_counts = {}

    def request(self, process):
        # look_windows = [(0, 0)]
        # for (output_name, output) in self.outputs.items():
        #     for output_proxy in output:
        #         look_windows.append(output_proxy.get_look_window())

        for (input_name, input) in self.inputs.items():
            input.request(process)

    def initialize(self, process):
        import harkmw.logic.mqtt_node
        import harkmw.types.native_node
        MQTTNode = harkmw.logic.mqtt_node.MQTTNode
        NativeNode = harkmw.types.native_node.NativeNode

        logger.debug('Node.initialize: look_window: in={}'.format(list(zip(
            [(tname, self.get_input_look_window(tname)) for tname in self.inputs.keys()]))))
        look_window = [max(x) for x in zip(
            *[self.get_input_look_window(tname) for tname in self.inputs.keys()])]
        logger.debug('Node.initialize: look_window={}'.format(look_window))

        for input in self.inputs.values():
            input.set_look_window(look_window)
            input.initialize(process)
            # iterator 入力端子を、count 番号管理の対象としない
            if self.network.network_type != 'iterator' or self != self.network.input_collector:
                self.input_counts[input] = -1

        if len(self.inputs) == 0:
            if self.process_name == process.process_name and self.network.process_name != process.process_name:
                process.subscribe(self.get_topic() + '/_INPUT', self, self)

        for name, output in self.outputs.items():
            for output_proxy in output:
                self.output_counts[output_proxy] = -1
                logger.debug(
                    'Node.Initialize: output_counts: self={}, {}'.format(
                        self.name, output_proxy))

        if isinstance(self, MQTTNode):
            for name, output in self.outputs.items():
                for o in output:
                    if isinstance(o, NativeNode) and o.type != 'Collector':
                        pass
                        # process.subscribe(self.get_topic() + '/' + name + '/_OUTPUT', self.network, self)
        logger.debug(
            'Node.Initialize: node: name={}, type={}, process={}'.format(
                self.name, self.type, self.process_name))
        logger.debug('  - Inputs:')
        for iname, input in self.inputs.items():
            logger.debug('    - {}: src-node={}, lookWindow={}'.format(iname,
                                                                       input.input_node.name, input.get_look_window()))
        logger.debug('  - Outputs:')
        for oname, output in self.outputs.items():
            for o in output:
                logger.debug('    - {}: dst-node={}, lookWindow={}'.format(oname,
                                                                           o.output_node.name, o.get_look_window()))

    def release(self):
        pass

    def is_input_ready(self, count):
        """ ループ count のデータを生成するための入力データは整っているか？
        """
        return True  # 既定動作は常に True

    def push_data(self, input, count, process, internal=False, outer_count=-1):
        """
        ループ count のデータが input に入力された場合にコールされる
        処理された場合は True を返す。他の input が準備できないない場合など、処理されない場合は False を返す。
        """
        logger.debug(
            "Node.push_data: Try push data for Node [{}/{}], input={}, count [{}/{}], #inputs={}, #outputs={}, ({})".format(
                self.type, self.name, input, count, self.count, len(
                    self.inputs), len(
                    self.outputs), self))

        if input is not None:
            self.input_counts[input] = count
        target_count = max(-1, min([count] + [self.input_counts[i] - i.get_look_window()[
                           0] for i in self.inputs.values() if i in self.input_counts]))
        if self.count >= target_count:
            logger.debug(
                'Node.push_data: returns without data processing: count={}, self.count={}, target_count={}'.format(
                    count, self.count, target_count))
            return self.count

        logger.debug(
            "Node.push_data: target-count: {} ({})".format(target_count, self.input_counts.items()))

        for i in self.inputs.values():
            logger.debug(
                "Node.push_data: Input[{}]: count={},{},{}, look_window={}".format(
                    i, i.count, self.count, count, i.get_look_window()))

        if self.name[-7:] == '_output':
            self.input_count = target_count
        self.count = target_count

        minimum_target_count = target_count
        for (output_name, output) in self.outputs.items():
            logger.debug("Node.push_data: output for {}: {}".format(
                self.name, output_name))
            if self.name[-7:] == '_output':
                logger.debug(
                    'Node.push_data: process: for output: start: output_name={}, target_count={}'.format(
                        output_name, target_count))
                self.process(
                    output_name,
                    target_count,
                    input_count=target_count,
                    output_count=outer_count)
                logger.debug(
                    'Node.push_data: process: for output: end: output_name={}, target_count={}'.format(
                        output_name, target_count))
            elif self.name[-6:] == '_input':
                logger.debug(
                    'Node.push_data: process: for input: start: output_name={}, target_count={}'.format(
                        output_name, target_count))
                self.process(
                    output_name,
                    target_count,
                    input_count=outer_count,
                    output_count=target_count)
                logger.debug(
                    'Node.push_data: process: for input: end: output_name={}, target_count={}'.format(
                        output_name, target_count))
            else:
                logger.debug(
                    'Node.push_data: process: start: output_name={}, target_count={}'.format(
                        output_name, target_count))
                self.process(output_name, target_count)
                logger.debug(
                    'Node.push_data: process: end: output_name={}, target_count={}'.format(
                        output_name, target_count))

            if not (
                    self.name[-7:] == '_output'and self.network.network_type == 'iterator'):
                for output_proxy in output:
                    logger.debug("Node.push_data: Push: {}: {}/{}: {}".format(
                        self.name, count, target_count, output_proxy))
                    # look_ahead = output_proxy.get_look_window()[0]
                    # diff_look_ahead = maximum_look_ahead - look_ahead

                    if output_proxy not in self.output_counts:
                        logger.error('output_proxy:  self={}, {}'.format(
                            self.name, output_proxy))
                        logger.error('not in keys: {}'.format(
                            self.output_counts.keys()))
                    # logger.debug(
                    #     'Node.push_data: output_count={}, target_count={}, look_ahead={}, diff_look_ahead={}'.format(
                    # self.output_counts[output_proxy], target_count,
                    # look_ahead, diff_look_ahead))

                    if (self.network.has_remote_nodes(
                            process) and output_proxy.output_node.process_name != process.process_name):

                        logger.debug(
                            'network_type={}, type={}'.format(
                                self.network.network_type,
                                output_proxy.output_node.type))
                        if (self.network.network_type != 'iterator' and output_proxy.output_node ==
                                self.network.output_collector):
                            pass
                        else:
                            logger.debug(
                                'Node.push_data: publishing: {}, {}, {}'.format(
                                    output_proxy.output_node.process_name,
                                    self.process_name,
                                    self.network.has_remote_nodes(process)))
                            t = self.get_topic() + "/" + output_proxy.input_terminal_name + "/_OUTPUT"

                            process.subscribe(t + '/ack', None)
                            process.publish(
                                t, target_count, self.get_output(
                                    output_proxy.input_terminal_name, target_count))

                            # wait for response
                            process.wait_topic(t + '/ack')
                            process.unsubscribe(t + '/ack', self)

                    self.output_counts[output_proxy] = target_count
                    minimum_target_count = min(
                        minimum_target_count, output_proxy.push_data(
                            self.output_counts[output_proxy], process))

        logger.debug(
            'Node.push_data: Return from Node.push_data: with minimum_target_count={}, node={}'.format(
                minimum_target_count, self.name))
        return minimum_target_count

    def get_output(self, output_name, count):
        raise NotImplementedError(
            'Node.get_output: name={}, output_name={}, count={}'.format(
                self.name, output_name, count))

    def get_input_look_window(self):
        raise NotImplementedError("Node.get_input_look_window")

    def set_input_look_window(self, input_id, lookWindow):
        raise NotImplementedError("Node.set_input_look_window")

    def get_output_look_window(self, output_name):
        raise NotImplementedError("Node.get_output_look_window")

    def set_output_look_window(self, output_id, lookWindow):
        raise NotImplementedError("Node.set_output_look_window")

    def get_input_count(self):
        raise NotImplementedError("")

    def get_output_count(self):
        raise NotImplementedError("")

    def get_data(self, terminal_name, count):
        raise NotImplementedError("")

    def get_delay(self):
        delay = 0
        for input in self.inputs.values():
            delay = max(delay, input.get_delay())
        return delay

    def get_topic(self):
        if self.topic is None:
            return "%s/%s" % (self.network.get_topic(), self.name)
        else:
            return self.topic

# end of file
